from collections import defaultdict
from typing import overload, Union, Callable, List

from crosstream.readers import CSVReader,ODBCReader

class HashJoin:
    def __init__(self):
        self.hash_buckets = defaultdict(list)

    def __build_hash_input(self, row, join_column_indexes):
        join_key = self.__build_join_key(row,join_column_indexes)
        self.hash_buckets[join_key].append(row)

    def __build_join_key(self,row,indices):
        # calculate the hash of join values
        join_values = []
        for col_index in indices:
            # convert row[col_index] to string because CSVs can't don't define numeric datatypes - would have to interpret them.
            join_values.append(str(hash(str(row[col_index]))))
        join_key = ''.join(join_values)

        return join_key

    def __process_matched_hashes(self,bucket_row,probe_row, bucket_join_column_indexes, probe_join_column_indexes):
        return tuple(bucket_row),tuple(probe_row,)

    reader = Union[CSVReader,ODBCReader]
    def inner_join(self, input_1: reader, input_2: reader, override_build_join_key=None, override_process_matched_hashes=None):

        if override_build_join_key != None:
            self.__build_join_key = override_build_join_key
        if override_process_matched_hashes != None:
            self.__process_matched_hashes = override_process_matched_hashes

        self.hash_buckets = defaultdict(list)

        for row in input_1:
            self.__build_hash_input(row,input_1.join_column_indexes)

        for row in input_2:
            join_key = self.__build_join_key(row,input_2.join_column_indexes)
            for bucket_row in self.hash_buckets[join_key]:
                record = self.__process_matched_hashes(bucket_row,row, input_1.join_column_indexes, input_2.join_column_indexes)
                if record != None:
                    yield record


import csv
from typing import List, overload
import pyodbc

class CSVReader():
    @overload
    def __init__(self, csv_file_path: str, has_headers: bool, join_columns: List[str], encoding: str):
        ...
    
    @overload
    def __init__(self, csv_file_path: str, has_headers: bool, join_columns: List[int], encoding: str):
        ...

    def __init__(self, csv_file_path, has_headers, join_columns, encoding='utf-8'):
        self.csv_file_path = csv_file_path
        self.has_headers=has_headers
        self.encoding=encoding

        if self.has_headers:
             with open(csv_file_path, encoding=self.encoding) as f:
                reader = csv.reader(f)
                headers = next(reader)
                self.column_names = headers

        if isinstance(join_columns, list) and all(isinstance(x, int) for x in join_columns):
            self.join_column_indexes = join_columns

        elif isinstance(join_columns, list) and all(isinstance(x, str) for x in join_columns):
            if has_headers==False:
                raise ValueError('has_headers=False, but strings passed to join_columns')

            # determine column indexes from names
            with open(csv_file_path, encoding=self.encoding) as f:
                reader = csv.reader(f)
                headers = next(reader)
                join_column_indexes=[]

                for column_name in join_columns:
                    column_index = headers.index(column_name)
                    join_column_indexes.append(column_index)
            
            self.join_column_indexes = join_column_indexes
        else:
            raise ValueError('join_columns must be of type List[int] or List[str]')
    
    def __iter__(self):
        with open(self.csv_file_path, encoding=self.encoding) as f:
            reader = csv.reader(f)

            # skip reading in the header row
            if self.has_headers:
                next(reader)

            for row in reader:
                yield tuple(row)
    
class ODBCReader():
    @overload
    def __init__(self, connection_string: str, query: str, join_columns: List[str]):
        ...
    
    @overload
    def __init__(self, connection_string: str, query: str, join_columns: List[int]):
        ...

    def __init__(self, connection_string, query, join_columns):
        self.connection_string = connection_string
        self.query = query

        cnxn = pyodbc.connect(self.connection_string)
        cursor = cnxn.cursor()
        cursor.execute(query)
        headers = [column[0] for column in cursor.description]
        self.column_names = headers

        if isinstance(join_columns, list) and all(isinstance(x, int) for x in join_columns):
            self.join_column_indexes = join_columns

        elif isinstance(join_columns, list) and all(isinstance(x, str) for x in join_columns):
            # determine column indexes from names
            cnxn = pyodbc.connect(self.connection_string)
            cursor = cnxn.cursor()
            cursor.execute(query)
            join_column_indexes=[]
            columns = [column[0] for column in cursor.description]
            
            for column_name in join_columns:
                column_index = columns.index(column_name)
                join_column_indexes.append(column_index)
            
            self.join_column_indexes = join_column_indexes
        else:
            raise ValueError('Must be of type List[int] or List[str]')
    
    def __iter__(self):
        connection = pyodbc.connect(self.connection_string)
        with connection:
            cursor = connection.cursor()
            cursor.execute(self.query)
            row = cursor.fetchone() 
            while row: 
                yield row
                row = cursor.fetchone()


from typing import overload, Union, Callable, List
from crosstream.readers import CSVReader, ODBCReader
from crosstream.hash_join import HashJoin

@overload
def read_csv(csv_file_path: str, has_headers: bool, join_columns: List[str], encoding: str):
    ...

@overload
def read_csv(csv_file_path: str, has_headers: bool, join_columns: List[int], encoding: str):
    ...

def read_csv(csv_file_path, has_headers, join_columns, encoding='utf-8'):
    """Lazily load the contents of a csv file one row at a time as needed."""

    return CSVReader(csv_file_path,has_headers,join_columns,encoding)


@overload
def read_odbc(connection_string: str, query: str, join_columns: List[str]):
    ...

@overload
def read_odbc(connection_string: str, query: str, join_columns: List[int]):
    ...

def read_odbc(connection_string, query, join_columns):
    """Lazily load the contents of an ODBC sql query one row at a time as needed."""
    return ODBCReader(connection_string, query, join_columns)

reader = Union[CSVReader,ODBCReader]
def inner_hash_join(input_1: reader, input_2: reader, override_build_join_key=None, override_process_matched_hashes=None):
    """Join two datasets using a hash join.

        The smaller dataset should be passed into input_1.
        """
    h=HashJoin()
    return h.inner_join(input_1,input_2,override_build_join_key,override_process_matched_hashes)


import pytest

import crosstream as cs
from crosstream.hash_join import HashJoin
import os, csv

def test_csv_false_headers(csv_input_data):
    with pytest.raises(ValueError):
        c1=cs.read_csv(csv_input_data[0],False,['col1','col2']) 
    
def test_csv_mixed_headers(csv_input_data):
    with pytest.raises(ValueError):
        c1=cs.read_csv(csv_input_data[0],True,[1,'col2']) 

def test_odbc_mixed_headers(sqlite_input_data):
    with pytest.raises(ValueError):
        q1=cs.read_odbc(sqlite_input_data,'SELECT * from test_data',[1,'col2'])

def test_csv_to_csv(tmp_path_factory,csv_input_data):
    # columns referred to by index and name
    c1=cs.read_csv(csv_input_data[0],True,[0,1])
    c2=cs.read_csv(csv_input_data[1], True, ['col1','col2'])

    h=HashJoin()
    
    output_file = os.path.join(tmp_path_factory.getbasetemp(),'csv_to_csv_output.csv')

    with open(output_file, 'w', newline='') as f:
        w =csv.writer(f)
        
        # write header column names
        w.writerow(c1.column_names + c2.column_names)

        for row_left,row_right in h.inner_join(c1,c2):
            # write matched results
            w.writerow(row_left + row_right)

    with open(output_file, 'r') as f:
        file_content = f.read()

        assert file_content == '''col1,col2,col1,col2
a,1,a,1
a,3,a,3
b,1,b,1
c,3,c,3
a1,1,a1,1
a1,1,a1,1
'''

def test_csv_to_odbc(tmp_path_factory,csv_input_data,sqlite_input_data):

    # columns referred to by index and name
    c1=cs.read_csv(csv_input_data[0],True,[0,1])
    q1=cs.read_odbc(sqlite_input_data,'SELECT * from test_data',['col1','col2'])

    output_file = os.path.join(tmp_path_factory.getbasetemp(),'csv_to_odbc_output.csv')

    with open(output_file, 'w', newline='') as f:
        w =csv.writer(f)
        
        # write header column names
        w.writerow(c1.column_names + q1.column_names)

        for row_left,row_right in cs.inner_hash_join(c1,q1):
            # write matched results
            w.writerow(row_left + row_right)

    with open(output_file, 'r') as f:
        file_content = f.read()

        assert file_content == '''col1,col2,col1,col2
a,1,a,1
a,3,a,3
b,1,b,1
c,3,c,3
a1,1,a1,1
a1,1,a1,1
'''

def test_odbc_to_odbc(tmp_path_factory,sqlite_input_data):

    # columns referred to by index and name
    q1=cs.read_odbc(sqlite_input_data,'SELECT * from test_data',[0])
    q2=cs.read_odbc(sqlite_input_data,'SELECT col2,col1 from test_data',[1])

    output_file = os.path.join(tmp_path_factory.getbasetemp(),'odbc_to_odbc_output.csv')

    with open(output_file, 'w', newline='') as f:
        w =csv.writer(f)
        
        # write header column names
        w.writerow(q1.column_names + q2.column_names)

        for row_left,row_right in cs.inner_hash_join(q1,q2):
            # write matched results
            w.writerow(row_left + row_right)

    with open(output_file, 'r') as f:
        file_content = f.read()
        print(file_content)
        assert file_content == '''col1,col2,col2,col1
a,1,1,a
a,3,1,a
a,1,3,a
a,3,3,a
b,1,1,b
c,3,3,c
d,1,1,d
d,2,1,d
d,1,2,d
d,2,2,d
a1,1,1,a1
'''

def test_custom_overrides(tmp_path_factory,csv_input_data,sqlite_input_data):

    # columns referred to by index and name
    c1=cs.read_csv(csv_input_data[0],True,[0,1])
    c2=cs.read_csv(csv_input_data[1], True, ['col1','col2'])

    # define a function for joining on criteria that is modified before insert into hash table
    def custom_join_key(row,indices):
        # calculate the hash of join values
        join_values = []
        for col_index in indices:
            # here we transform our join key, removing any spaces from our values
            join_values.append(str(hash(str(row[col_index]).replace(' ',''))))
        join_key = ''.join(join_values)

        return join_key

    # define a function for performing additional transformations or adding additional outputs before the columns are returned
    def custom_process_matched_hashes(bucket_row,probe_row, bucket_join_column_indexes, probe_join_column_indexes):
        # adding a new column indicating the weights of these matches are equal to 1
        weight=1.0
        return tuple(bucket_row),tuple(probe_row),(weight,)

    output_file = os.path.join(tmp_path_factory.getbasetemp(),'csv_to_odbc_overrides_output.csv')

    with open(output_file, 'w', newline='') as f:
        w =csv.writer(f)
        
        # write header column names
        w.writerow(c1.column_names + c2.column_names + ['weight'])

        for row_left,row_right,weight in cs.inner_hash_join(c1,c2,override_build_join_key=custom_join_key,override_process_matched_hashes=custom_process_matched_hashes):
            # write matched results
            w.writerow(row_left + row_right + weight)

    with open(output_file, 'r') as f:
        file_content = f.read()

        assert file_content == '''col1,col2,col1,col2,weight
a,1,a,1,1.0
a,3,a,3,1.0
b,1,b,1,1.0
c,3,c,3,1.0
d e,1,de,1,1.0
a1,1,a1,1,1.0
a1,1,a1,1,1.0
'''

if __name__ == "__main__":
    pytest.main(["tests/test_hash_join.py", "-s", "-k", "test_csv_to_odbc"])

